# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

import abc
import six


@six.add_metaclass(abc.ABCMeta)
class AuthenticationProvider(object):
    """Super class for all providing known types of authentication mechanism like
    x509 and SAS based authentication.

    :ivar str hostname: Hostname
    :ivar str device_id: Device ID
    :ivar str module_id: Module ID
    """

    def __init__(self, hostname, device_id, module_id=None):
        """Initializer for AuthenticationProvider

        :param str hostname: Hostname
        :param str device_id: Device ID
        :param str module_id: Module ID (optional)
        """
        self.hostname = hostname
        self.device_id = device_id
        self.module_id = module_id

    @abc.abstractmethod
    def get_current_sas_token(self):
        """Get the current SharedAuthenticationSignature string.
        """
        pass

    # @abc.abstractstaticmethod
    @abc.abstractmethod
    def parse(source):
        """Method needs to be implemented as static method in child authentications providers.
        :param:source The source in string. This could be connections string or a shared access signature string.
        """
        pass



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""This module contains tools for adapting sync code for use in async coroutines."""

import functools
import azure.iot.device.common.asyncio_compat as asyncio_compat


def emulate_async(fn):
    """Apply as a decorator to emulate async behavior with a sync function/method
    via usage of multithreading.

    :param fn: The sync function to be run in async.
    :returns: A coroutine function that will call the given sync function.
    """

    @functools.wraps(fn)
    async def async_fn_wrapper(*args, **kwargs):
        loop = asyncio_compat.get_running_loop()

        # Run fn in default ThreadPoolExecutor (CPU * 5 threads)
        return await loop.run_in_executor(None, functools.partial(fn, *args, **kwargs))

    return async_fn_wrapper


class AwaitableCallback(object):
    """A sync callback whose completion can be waited upon.
    """

    def __init__(self, callback):
        """Creates an instance of an AwaitableCallback from a callback function.

        :param callback: Callback function to be made awaitable.
        """
        loop = asyncio_compat.get_running_loop()
        self.future = asyncio_compat.create_future(loop)

        def wrapping_callback(*args, **kwargs):
            result = callback(*args, **kwargs)
            # Use event loop from outer scope, since the threads it will be used in will not have
            # an event loop. future.set_result() has to be called in an event loop or it does not work.
            loop.call_soon_threadsafe(self.future.set_result, result)
            return result

        self.callback = wrapping_callback

    def __call__(self, *args, **kwargs):
        """Calls the callback. Returns the result.
        """
        return self.callback(*args, **kwargs)

    async def completion(self):
        """Awaitable coroutine method that will return once the AwaitableCallback
        has been completed.

        :returns: Result of the callback when it was called.
        """
        return await self.future



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

import logging
from datetime import date
import six.moves.urllib as urllib
import six.moves.queue as queue
from .mqtt_provider import MQTTProvider
from transitions import Machine
from azure.iot.device.iothub.transport.abstract_transport import AbstractTransport
from azure.iot.device.iothub.transport import constant
from azure.iot.device.iothub.models import Message


"""
The below import is for generating the state machine graph.
"""
# from transitions.extensions import LockedGraphMachine as Machine

logger = logging.getLogger(__name__)

"""
A note on names, design, and code flow:

This transport is a state machine which is responsible for coordinating
several different things.  (I would like to say that it coordinates several
events, but the word "event" is very overloaded, especially in this context,
so I hesitate to add one more overload).

In particular, it needs to coordinate external things:
    1. Things the caller wants to do, such as "connect", "send message", etc.
    2. Calls into the transport provider
    3. Completion callbacks from the transport provider.
    4. Completion callbacks from this object into the caller.

and also internal things:
    5. The "state" of transport (connected, disconnected, etc).
    6. Transitions between possible states.

Since one caller-initiated action results in many different things happening, this
class uses the following conventions:

1. Actions that the caller can initiate are all named without an underscore at the beginning,
    such as "connect", "send_message", etc.

2. All internal functions are prefixed with an underscore.  This is "the pythonic way", but it bears repeating.
    Internal functions are internal and should be called by external code.

2. Actions will typically trigger a state machine event.  State machine triggers, wihch may
    or may not change state, are all prefixed with _trig_ (_trig_connect, _trig_add_pending_action_to_queue, etc)
    The "_trig" indicates that this is a operation on the state machine.  _trig_* functions are also unusual in
    that they get added to the transport object at runtime when intiializing the Machine object.

3. Functions which call into the provider are prefixed with "_call_provider_", such as _call_provider_connect.
    These are always called as part of state machine transitions.  Calls from the caller should not go directly into the
    provider without going through the state machine.

4. When the provider completes an action or receives an acknowledgement, it will call back into this object using
    functions which are all prefixed with "_on_provider_", such as "_on_provider_connect_complete".  These callback functions
    will, most of the time, trigger additional state machine transitions by calling _trig_ functions.

5. Functions which are called by the state machine as side-effects of state transitions will accept event_data objects
    as the second parameter (after self).  The event_data structure contains information about the trigger or the transition
    which caused the side-effect.

6. Callbacks from this object into caller code, when not passed in as `callback` parameters to function calls, are prefixed with
    on_ (with no underscore), such as "on_transport_connected'.  Because most callbacks are passed in as function parameters,
    there are very few callbacks like this.

"""


TOPIC_POS_DEVICE = 4
TOPIC_POS_MODULE = 6
TOPIC_POS_INPUT_NAME = 5


class TransportAction(object):
    """
    base class representing various actions that can be taken
    when the transport is connected.  When the MqttTransport user
    calls a function that requires the transport to be connected,
    a TransportAction object is created and added to the action
    queue.  Then, when the transport is actually connected, it
    loops through the objects in the action queue and executes them
    one by one.
    """

    def __init__(self, callback):
        self.callback = callback


class SendMessageAction(TransportAction):
    """
    TransportAction object used to send a telemetry message or an
    output message
    """

    def __init__(self, message, callback):
        TransportAction.__init__(self, callback)
        self.message = message


class SubscribeAction(TransportAction):
    """
    TransportAction object used to subscribe to a specific MQTT topic
    """

    def __init__(self, topic, qos, callback):
        TransportAction.__init__(self, callback)
        self.topic = topic
        self.qos = qos


class UnsubscribeAction(TransportAction):
    """
    TransportAction object used to unsubscribe from a specific MQTT topic
    """

    def __init__(self, topic, callback):
        TransportAction.__init__(self, callback)
        self.topic = topic


class MethodReponseAction(TransportAction):
    """
    TransportAction object used to send a method response back to the service.
    """

    def __init__(self, method_response, callback):
        TransportAction.__init__(self, callback)
        self.method_response = method_response


class MQTTTransport(AbstractTransport):
    def __init__(self, auth_provider):
        """
        Constructor for instantiating a transport
        :param auth_provider: The authentication provider
        """
        AbstractTransport.__init__(self, auth_provider)
        self.topic = self._get_telemetry_topic_for_publish()
        self._mqtt_provider = None

        # Queue of actions that will be executed once the transport is connected.
        # Currently, we use a queue, which is FIFO, but the actual order doesn't matter
        # since each action stands on its own.
        self._pending_action_queue = queue.Queue()

        # Object which maps mid->callback for actions which are in flight.  This is
        # used to call back into the caller to indicate that an action is complete.
        self._in_progress_actions = {}

        # Map of responses we receive with a MID that is not in the _in_progress_actions map.
        # We need this because sometimes a SUBSCRIBE or a PUBLISH will complete before the call
        # to subscribe() or publish() returns.
        self._responses_with_unknown_mid = {}

        self._connect_callback = None
        self._disconnect_callback = None

        self._c2d_topic = None
        self._input_topic = None

        states = ["disconnected", "connecting", "connected", "disconnecting"]

        transitions = [
            {
                "trigger": "_trig_connect",
                "source": "disconnected",
                "dest": "connecting",
                "after": "_call_provider_connect",
            },
            {"trigger": "_trig_connect", "source": ["connecting", "connected"], "dest": None},
            {
                "trigger": "_trig_provider_connect_complete",
                "source": "connecting",
                "dest": "connected",
                "after": "_execute_actions_in_queue",
            },
            {
                "trigger": "_trig_disconnect",
                "source": ["disconnected", "disconnecting"],
                "dest": None,
            },
            {
                "trigger": "_trig_disconnect",
                "source": "connected",
                "dest": "disconnecting",
                "after": "_call_provider_disconnect",
            },
            {
                "trigger": "_trig_provider_disconnect_complete",
                "source": "disconnecting",
                "dest": "disconnected",
            },
            {
                "trigger": "_trig_add_action_to_pending_queue",
                "source": "connected",
                "before": "_add_action_to_queue",
                "dest": None,
                "after": "_execute_actions_in_queue",
            },
            {
                "trigger": "_trig_add_action_to_pending_queue",
                "source": "connecting",
                "before": "_add_action_to_queue",
                "dest": None,
            },
            {
                "trigger": "_trig_add_action_to_pending_queue",
                "source": "disconnected",
                "before": "_add_action_to_queue",
                "dest": "connecting",
                "after": "_call_provider_connect",
            },
            {
                "trigger": "_trig_on_shared_access_string_updated",
                "source": "connected",
                "dest": "connecting",
                "after": "_call_provider_reconnect",
            },
            {
                "trigger": "_trig_on_shared_access_string_updated",
                "source": ["disconnected", "disconnecting"],
                "dest": None,
            },
        ]

        def _on_transition_complete(event_data):
            if not event_data.transition:
                dest = "[no transition]"
            else:
                dest = event_data.transition.dest
            logger.info(
                "Transition complete.  Trigger=%s, Dest=%s, result=%s, error=%s",
                event_data.event.name,
                dest,
                str(event_data.result),
                str(event_data.error),
            )

        self._state_machine = Machine(
            model=self,
            states=states,
            transitions=transitions,
            initial="disconnected",
            send_event=True,  # This has nothing to do with telemetry events.  This tells the machine use event_data structures to hold transition arguments
            finalize_event=_on_transition_complete,
            queued=True,
        )

        # to render the state machine as a PNG:
        # 1. apt install graphviz
        # 2. pip install pygraphviz
        # 3. change import line at top of this file to import LockedGraphMachine as Machine
        # 4. uncomment the following line
        # 5. run this code
        # self.get_graph().draw('mqtt_transport.png', prog='dot')

        self._create_mqtt_provider()

    def _call_provider_connect(self, event_data):
        """
        Call into the provider to connect the transport.

        This is called by the state machine as part of a state transition

        :param EventData event_data:  Object created by the Transitions library with information about the state transition
        """
        logger.info("Calling provider connect")
        password = self._auth_provider.get_current_sas_token()
        self._mqtt_provider.connect(password)

        if hasattr(self._auth_provider, "token_update_callback"):
            self._auth_provider.token_update_callback = self._on_shared_access_string_updated

    def _call_provider_disconnect(self, event_data):
        """
        Call into the provider to disconnect the transport.

        This is called by the state machine as part of a state transition

        :param EventData event_data:  Object created by the Transitions library with information about the state transition
        """
        logger.info("Calling provider disconnect")
        self._mqtt_provider.disconnect()
        self._auth_provider.disconnect()

    def _call_provider_reconnect(self, event_data):
        """
        Call into the provider to reconnect the transport.

        This is called by the state machine as part of a state transition

        :param EventData event_data:  Object created by the Transitions library with information about the state transition
        """
        password = self._auth_provider.get_current_sas_token()
        self._mqtt_provider.reconnect(password)

    def _on_provider_connect_complete(self):
        """
        Callback that is called by the provider when the connection has been established
        """
        logger.info("_on_provider_connect_complete")
        self._trig_provider_connect_complete()

        if self.on_transport_connected:
            self.on_transport_connected("connected")
        callback = self._connect_callback
        if callback:
            self._connect_callback = None
            callback()

    def _on_provider_disconnect_complete(self):
        """
        Callback that is called by the provider when the connection has been disconnected
        """
        logger.info("_on_provider_disconnect_complete")
        self._trig_provider_disconnect_complete()

        if self.on_transport_disconnected:
            self.on_transport_disconnected("disconnected")
        callback = self._disconnect_callback
        if callback:
            self._disconnect_callback = None
            callback()

    def _on_provider_publish_complete(self, mid):
        """
        Callback that is called by the provider when it receives a PUBACK from the service

        :param mid: message id that was returned by the provider when `publish` was called.  This is used to tie the
            PUBLISH to the PUBACK.
        """
        if mid in self._in_progress_actions:
            callback = self._in_progress_actions[mid]
            del self._in_progress_actions[mid]
            callback()
        else:
            logger.warning("PUBACK received with unknown MID: %s", str(mid))
            self._responses_with_unknown_mid[
                mid
            ] = mid  # storing MID for now.  will probably store result code later.

    def _on_provider_subscribe_complete(self, mid):
        """
        Callback that is called by the provider when it receives a SUBACK from the service

        :param mid: message id that was returned by the provider when `subscribe` was called.  This is used to tie the
            SUBSCRIBE to the SUBACK.
        """
        if mid in self._in_progress_actions:
            callback = self._in_progress_actions[mid]
            del self._in_progress_actions[mid]
            callback()
        else:
            logger.warning("SUBACK received with unknown MID: %s", str(mid))
            self._responses_with_unknown_mid[
                mid
            ] = mid  # storing MID for now.  will probably store result code later.

    def _on_provider_message_received_callback(self, topic, payload):
        """
        Callback that is called by the provider when a message is received.  This message can be any MQTT message,
        including, but not limited to, a C2D message, an input message, a TWIN patch, a twin response (/res), and
        a method invocation.  This function needs to decide what kind of message it is based on the topic name and
        take the correct action.

        :param topic: MQTT topic name that the message arrived on
        :param payload: Payload of the message
        """
        logger.info("Message received on topic %s", topic)
        message_received = Message(payload)
        # TODO : Discuss everything in bytes , need to be changed, specially the topic
        topic_str = topic.decode("utf-8")
        topic_parts = topic_str.split("/")

        if _is_input_topic(topic_str):
            input_name = topic_parts[TOPIC_POS_INPUT_NAME]
            message_received.input_name = input_name
            _extract_properties(topic_parts[TOPIC_POS_MODULE], message_received)
            self.on_transport_input_message_received(input_name, message_received)
        elif _is_c2d_topic(topic_str):
            _extract_properties(topic_parts[TOPIC_POS_DEVICE], message_received)
            self.on_transport_c2d_message_received(message_received)
        else:
            pass  # is there any other case

    def _on_provider_unsubscribe_complete(self, mid):
        """
        Callback that is called by the provider when it receives an UNSUBACK from the service

        :param mid: message id that was returned by the provider when `unsubscribe` was called.  This is used to tie the
            UNSUBSCRIBE to the UNSUBACK.
        """
        if mid in self._in_progress_actions:
            callback = self._in_progress_actions[mid]
            del self._in_progress_actions[mid]
            callback()
        else:
            logger.warning("UNSUBACK received with unknown MID: %s", str(mid))
            self._responses_with_unknown_mid[
                mid
            ] = mid  # storing MID for now.  will probably store result code later.

    def _add_action_to_queue(self, event_data):
        """
        Queue an action for running later.  All actions that need to run while connected end up in
        this queue, even if they're going to be run immediately.

        This is called by the state machine as part of a state transition

        :param EventData event_data:  Object created by the Transitions library with information about the state transition
        """
        self._pending_action_queue.put_nowait(event_data.args[0])

    def _execute_action(self, action):
        """
        Execute an action from the action queue.  This is called when the transport is connected and the
        state machine is able to execute individual actions.

        :param TransportAction action: object containing the details of the action to be executed
        """

        if isinstance(action, SendMessageAction):
            logger.info("running SendMessageAction")
            message_to_send = action.message
            encoded_topic = _encode_properties(
                message_to_send, self._get_telemetry_topic_for_publish()
            )
            mid = self._mqtt_provider.publish(encoded_topic, message_to_send.data)
            if mid in self._responses_with_unknown_mid:
                del self._responses_with_unknown_mid[mid]
                action.callback()
            else:
                self._in_progress_actions[mid] = action.callback

        elif isinstance(action, SubscribeAction):
            logger.info("running SubscribeAction topic=%s qos=%s", action.topic, action.qos)
            mid = self._mqtt_provider.subscribe(action.topic, action.qos)
            logger.info("subscribe mid = %s", mid)
            if mid in self._responses_with_unknown_mid:
                del self._responses_with_unknown_mid[mid]
                action.callback()
            else:
                self._in_progress_actions[mid] = action.callback

        elif isinstance(action, UnsubscribeAction):
            logger.info("running UnsubscribeAction")
            mid = self._mqtt_provider.unsubscribe(action.topic)
            if mid in self._responses_with_unknown_mid:
                del self._responses_with_unknown_mid[mid]
                action.callback()
            else:
                self._in_progress_actions[mid] = action.callback

        elif isinstance(action, MethodReponseAction):
            logger.info("running MethodResponseAction")
            topic = "TODO"
            mid = self._mqtt_provider.publish(topic, action.method_response)
            if mid in self._responses_with_unknown_mid:
                del self._responses_with_unknown_mid[mid]
                action.callback()
            else:
                self._in_progress_actions[mid] = action.callback

        else:
            logger.error("Removed unknown action type from queue.")

    def _execute_actions_in_queue(self, event_data):
        """
        Execute any actions that are waiting in the action queue.
        This is called by the state machine as part of a state transition.
        This function actually calls down into the provider to perform the necessary operations.

        :param EventData event_data:  Object created by the Transitions library with information about the state transition
        """
        logger.info("checking _pending_action_queue")
        while True:
            try:
                action = self._pending_action_queue.get_nowait()
            except queue.Empty:
                logger.info("done checking queue")
                return

            self._execute_action(action)

    def _create_mqtt_provider(self):
        """
        Create the provider object which is used by this instance to communicate with the service.
        No network communication can take place without a provider object.
        """
        client_id = self._auth_provider.device_id

        if self._auth_provider.module_id:
            client_id += "/" + self._auth_provider.module_id

        username = self._auth_provider.hostname + "/" + client_id + "/" + "?api-version=2018-06-30"

        hostname = None
        if hasattr(self._auth_provider, "gateway_hostname"):
            hostname = self._auth_provider.gateway_hostname
        if not hostname or len(hostname) == 0:
            hostname = self._auth_provider.hostname

        if hasattr(self._auth_provider, "ca_cert"):
            ca_cert = self._auth_provider.ca_cert
        else:
            ca_cert = None

        self._mqtt_provider = MQTTProvider(client_id, hostname, username, ca_cert=ca_cert)

        self._mqtt_provider.on_mqtt_connected = self._on_provider_connect_complete
        self._mqtt_provider.on_mqtt_disconnected = self._on_provider_disconnect_complete
        self._mqtt_provider.on_mqtt_published = self._on_provider_publish_complete
        self._mqtt_provider.on_mqtt_subscribed = self._on_provider_subscribe_complete
        self._mqtt_provider.on_mqtt_unsubscribed = self._on_provider_unsubscribe_complete
        self._mqtt_provider.on_mqtt_message_received = self._on_provider_message_received_callback

    def _get_topic_base(self):
        """
        return the string that is at the beginning of all topics for this
        device/module
        """

        if self._auth_provider.module_id:
            return (
                "devices/"
                + self._auth_provider.device_id
                + "/modules/"
                + self._auth_provider.module_id
            )
        else:
            return "devices/" + self._auth_provider.device_id

    def _get_telemetry_topic_for_publish(self):
        """
        return the topic string used to publish telemetry
        """
        return self._get_topic_base() + "/messages/events/"

    def _get_c2d_topic_for_subscribe(self):
        """
        :return: The topic for cloud to device messages.It is of the format
        "devices/<deviceid>/messages/devicebound/#"
        """
        return self._get_topic_base() + "/messages/devicebound/#"

    def _get_input_topic_for_subscribe(self):
        """
        :return: The topic for input messages. It is of the format
        "devices/<deviceId>/modules/<moduleId>/messages/inputs/#"
        """
        return self._get_topic_base() + "/inputs/#"

    def connect(self, callback=None):
        """
        Connect to the service.

        :param callback: callback which is called when the connection to the service is complete.
        """
        logger.info("connect called")
        self._connect_callback = callback
        self._trig_connect()

    def disconnect(self, callback=None):
        """
        Disconnect from the service.

        :param callback: callback which is called when the connection to the service has been disconnected
        """
        logger.info("disconnect called")
        self._disconnect_callback = callback
        self._trig_disconnect()

    def send_event(self, message, callback=None):
        """
        Send a telemetry message to the service.

        :param callback: callback which is called when the message publish has been acknowledged by the service.
        """
        action = SendMessageAction(message, callback)
        self._trig_add_action_to_pending_queue(action)

    def send_output_event(self, message, callback=None):
        """
        Send an output message to the service.

        :param callback: callback which is called when the message publish has been acknowledged by the service.
        """
        action = SendMessageAction(message, callback)
        self._trig_add_action_to_pending_queue(action)

    def send_method_response(self, method, payload, status, callback=None):
        raise NotImplementedError

    def _on_shared_access_string_updated(self):
        """
        Callback which is called by the authentication provider when the shared access string has been updated.
        """
        self._trig_on_shared_access_string_updated()

    def enable_feature(self, feature_name, callback=None):
        """
        Enable the given feature by subscribing to the appropriate topics.

        :param feature_name: one of the feature name constants from constant.py
        :param callback: callback which is called when the feature is enabled
        """
        logger.info("enable_feature %s called", feature_name)
        if feature_name == constant.INPUT_MSG:
            self._enable_input_messages(callback)
        elif feature_name == constant.C2D_MSG:
            self._enable_c2d_messages(callback)
        elif feature_name == constant.METHODS:
            self._enable_methods(callback)
        else:
            logger.error("Feature name {} is unknown".format(feature_name))
            raise ValueError("Invalid feature name")

    def disable_feature(self, feature_name, callback=None):
        """
        Disable the given feature by subscribing to the appropriate topics.
        :param callback: callback which is called when the feature is disabled

        :param feature_name: one of the feature name constants from constant.py
        """
        logger.info("disable_feature %s called", feature_name)
        if feature_name == constant.INPUT_MSG:
            self._disable_input_messages(callback)
        elif feature_name == constant.C2D_MSG:
            self._disable_c2d_messages(callback)
        elif feature_name == constant.METHODS:
            self._disable_methods(callback)
        else:
            logger.error("Feature name {} is unknown".format(feature_name))
            raise ValueError("Invalid feature name")

    def _enable_input_messages(self, callback=None):
        """
        Helper function to enable input messages

        :param callback: callback which is called when the feature is enabled
        """
        action = SubscribeAction(
            topic=self._get_input_topic_for_subscribe(), qos=1, callback=callback
        )
        self._trig_add_action_to_pending_queue(action)
        self.feature_enabled[constant.INPUT_MSG] = True

    def _disable_input_messages(self, callback=None):
        """
        Helper function to disable input messages

        :param callback: callback which is called when the feature is disabled
        """
        action = UnsubscribeAction(topic=self._get_input_topic_for_subscribe(), callback=callback)
        self._trig_add_action_to_pending_queue(action)
        self.feature_enabled[constant.INPUT_MSG] = False

    def _enable_c2d_messages(self, callback=None):
        """
        Helper function to enable c2de messages

        :param callback: callback which is called when the feature is enabled
        """
        action = SubscribeAction(
            topic=self._get_c2d_topic_for_subscribe(), qos=1, callback=callback
        )
        self._trig_add_action_to_pending_queue(action)
        self.feature_enabled[constant.C2D_MSG] = True

    def _disable_c2d_messages(self, callback=None):
        """
        Helper function to disabled c2d messages

        :param callback: callback which is called when the feature is disabled
        """
        action = UnsubscribeAction(topic=self._get_c2d_topic_for_subscribe(), callback=callback)
        self._trig_add_action_to_pending_queue(action)
        self.feature_enabled[constant.C2D_MSG] = False

    def _enable_methods(self, callback=None, qos=1):
        """
        Helper function to enable methods

        :param callback: callback which is called when the feature is enabled.
        :param qos: Quality of Serivce level
        """
        action = SubscribeAction(self._get_method_topic_for_subscribe(), qos, callback)
        self._trig_add_action_to_pending_queue(action)
        self.feature_enabled[constant.METHODS] = True

    def _disable_methods(self, callback=None):
        """
        Helper function to disable methods

        :param callback: callback which is called when the feature is disabled
        """
        action = UnsubscribeAction(self._get_method_topic_for_subscribe(), callback)
        self._trig_add_action_to_pending_queue(action)
        self.feature_enabled[constant.METHODS] = False


def _is_c2d_topic(split_topic_str):
    """
    Topics for c2d message are of the following format:
    devices/<deviceId>/messages/devicebound
    :param split_topic_str: The already split received topic string
    """
    if "messages/devicebound" in split_topic_str and len(split_topic_str) > 4:
        return True
    return False


def _is_input_topic(split_topic_str):
    """
    Topics for inputs are of the following format:
    devices/<deviceId>/modules/<moduleId>/messages/inputs/<inputName>
    :param split_topic_str: The already split received topic string
    """
    if "inputs" in split_topic_str and len(split_topic_str) > 6:
        return True
    return False


def _extract_properties(properties, message_received):
    """
    Extract key=value pairs from custom properties and set the properties on the received message.
    :param properties: The properties string which is ampersand(&) delimited key=value pair.
    :param message_received: The message received with the payload in bytes
    """
    key_value_pairs = properties.split("&")

    for entry in key_value_pairs:
        pair = entry.split("=")
        key = urllib.parse.unquote_plus(pair[0])
        value = urllib.parse.unquote_plus(pair[1])

        if key == "$.mid":
            message_received.message_id = value
        elif key == "$.cid":
            message_received.correlation_id = value
        elif key == "$.uid":
            message_received.user_id = value
        elif key == "$.to":
            message_received.to = value
        elif key == "$.ct":
            message_received.content_type = value
        elif key == "$.ce":
            message_received.content_encoding = value
        else:
            message_received.custom_properties[key] = value


def _encode_properties(message_to_send, topic):
    """
    uri-encode the system properties of a message as key-value pairs on the topic with defined keys.
    Additionally if the message has user defined properties, the property keys and values shall be
    uri-encoded and appended at the end of the above topic with the following convention:
    '<key>=<value>&<key2>=<value2>&<key3>=<value3>(...)'
    :param message_to_send: The message to send
    :param topic: The topic which has not been encoded yet. For a device it looks like
    "devices/<deviceId>/messages/events/" and for a module it looks like
    "devices/<deviceId>/modules/<moduleId>/messages/events/
    :return: The topic which has been uri-encoded
    """
    system_properties = {}
    if message_to_send.output_name:
        system_properties["$.on"] = message_to_send.output_name
    if message_to_send.message_id:
        system_properties["$.mid"] = message_to_send.message_id

    if message_to_send.correlation_id:
        system_properties["$.cid"] = message_to_send.correlation_id

    if message_to_send.user_id:
        system_properties["$.uid"] = message_to_send.user_id

    if message_to_send.to:
        system_properties["$.to"] = message_to_send.to

    if message_to_send.content_type:
        system_properties["$.ct"] = message_to_send.content_type

    if message_to_send.content_encoding:
        system_properties["$.ce"] = message_to_send.content_encoding

    if message_to_send.expiry_time_utc:
        system_properties["$.exp"] = (
            message_to_send.expiry_time_utc.isoformat()
            if isinstance(message_to_send.expiry_time_utc, date)
            else message_to_send.expiry_time_utc
        )

    system_properties_encoded = urllib.parse.urlencode(system_properties)
    topic += system_properties_encoded

    if message_to_send.custom_properties and len(message_to_send.custom_properties) > 0:
        topic += "&"
        user_properties_encoded = urllib.parse.urlencode(message_to_send.custom_properties)
        topic += user_properties_encoded

    return topic



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""This module contains a manager for inboxes."""

import logging

logger = logging.getLogger(__name__)


class InboxManager(object):
    """Manages the various Inboxes for a client.

    :ivar c2d_message_inbox: The C2D message Inbox.
    :ivar input_message_inboxes: A dictionary mapping input names to input message Inboxes.
    :ivar generic_method_request_inbox: The generic method request Inbox.
    :ivar named_method_request_inboxes: A dictionary mapping method names to method request Inboxes.
    """

    def __init__(self, inbox_type):
        """Initializer for the InboxManager.

        :param inbox_type: An Inbox class that the manager will use to create Inboxes.
        """
        self._create_inbox = inbox_type
        self.c2d_message_inbox = self._create_inbox()
        self.input_message_inboxes = {}
        self.generic_method_request_inbox = self._create_inbox()
        self.named_method_request_inboxes = {}

    def get_input_message_inbox(self, input_name):
        """Retrieve the input message Inbox for a given input.

        If the Inbox does not already exist, it will be created.

        :param str input_name: The name of the input for which the associated Inbox is desired.
        :returns: An Inbox for input messages on the selected input.
        """
        try:
            inbox = self.input_message_inboxes[input_name]
        except KeyError:
            # Create new Inbox for input if it does not yet exist
            inbox = self._create_inbox()
            self.input_message_inboxes[input_name] = inbox

        return inbox

    def get_c2d_message_inbox(self):
        """Retrieve the Inbox for C2D messages.

        :returns: An Inbox for C2D messages.
        """
        return self.c2d_message_inbox

    def get_method_request_inbox(self, method_name=None):
        """Retrieve the method request Inbox for a given method name if provided,
        or for generic method requests if not.

        If the Inbox does not already exist, it will be created.

        :param str method_name: Optional. The name of the method for which the
        associated Inbox is desired.
        :returns: An Inbox for method requests.
        """
        if method_name:
            try:
                inbox = self.named_method_request_inboxes[method_name]
            except KeyError:
                # Create a new Inbox for the method name
                inbox = self._create_inbox()
                self.named_method_request_inboxes[method_name] = inbox
        else:
            inbox = self.generic_method_request_inbox

        return inbox

    def clear_all_method_requests(self):
        """Delete all method requests currently in inboxes.
        """
        self.generic_method_request_inbox.clear()
        for inbox in self.named_method_request_inboxes.values():
            inbox.clear()

    def route_input_message(self, input_name, incoming_message):
        """Route an incoming input message to the correct input message Inbox.

        If the input is unknown, the message will be dropped.

        :param str input_name: The name of the input to route the message to.
        :param incoming_message: The message to be routed.

        :returns: Boolean indicating if message was successfuly routed or not.
        """
        try:
            inbox = self.input_message_inboxes[input_name]
        except KeyError:
            logger.warning("No input message inbox for {} - dropping message".format(input_name))
            return False
        else:
            inbox._put(incoming_message)
            logger.info("Input message sent to {} inbox".format(input_name))
            return True

    def route_c2d_message(self, incoming_message):
        """Route an incoming C2D message to the C2D message Inbox.

        :param incoming_message: The message to be routed.

        :returns: Boolean indicating if message was successfully routed or not.
        """
        self.c2d_message_inbox._put(incoming_message)
        logger.info("C2D message sent to inbox")
        return True

    def route_method_request(self, incoming_method_request):
        """Route an incoming method request to the correct method request Inbox.

        If the method name is recognized, it will be routed to a method-specific Inbox.
        Otherwise, it will be routed to the generic method request Inbox.

        :param incoming_method_request: The method request to be routed.

        :returns: Boolean indicating if the method request was successfully routed or not.
        """
        try:
            inbox = self.named_method_request_inboxes[incoming_method_request.name]
        except KeyError:
            inbox = self.generic_method_request_inbox
        inbox._put(incoming_method_request)
        return True



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

import logging
from .authentication_provider import AuthenticationProvider

"""
The urllib, urllib2, and urlparse modules from Python 2 have been combined in the urllib package in Python 3
The six.moves.urllib package is a python version-independent location of the above functionality.
"""
import six.moves.urllib as urllib

logger = logging.getLogger(__name__)

URI_SEPARATOR = "/"
DELIMITER = "&"
VALUE_SEPARATOR = "="
PARTS_SEPARATOR = " "

SIGNATURE = "sig"
SHARED_ACCESS_KEY_NAME = "skn"
RESOURCE_URI = "sr"
EXPIRY = "se"

_valid_keys = [SIGNATURE, SHARED_ACCESS_KEY_NAME, RESOURCE_URI, EXPIRY]


class SharedAccessSignatureAuthenticationProvider(AuthenticationProvider):
    """
    The Shared Access Signature Authentication Provider.
    This provider already contains the sas token which will be needed to authenticate with The IoT hub.
    """

    def __init__(self, hostname, device_id, module_id, sas_token_str):
        """
        Constructor for Shared Access Signature Authentication Provider
        """
        logger.info("Using SAS authentication for {%s, %s, %s}", hostname, device_id, module_id)
        AuthenticationProvider.__init__(self, hostname, device_id, module_id)
        self.sas_token_str = sas_token_str

    def get_current_sas_token(self):
        """
        :return: the string representation of the current Shared Access Signature
        """
        return self.sas_token_str

    @staticmethod
    def parse(sas_token_str):
        """
        This method creates a Shared Access Signature Authentication Provider from a string, and sets properties for each of the parsed
        fields in the string. Also validates the required properties of the shared access signature.
        :param sas_token_str: The ampersand-delimited string of 'name=value' pairs.
        The input may look like the following formations:-
        SharedAccessSignature sr=<resource_uri>&sig=<signature>&se=<expiry>
        SharedAccessSignature sr=<resource_uri>&sig=<signature>&skn=<keyname>&se=<expiry>
        :return: The Shared Access Signature Authentication Provider constructed
        """
        try:
            parts = sas_token_str.split(PARTS_SEPARATOR)
            sas_args = parts[1].split(DELIMITER)
            d = dict(arg.split(VALUE_SEPARATOR, 1) for arg in sas_args)
        except (IndexError, ValueError, AttributeError):
            raise ValueError(
                "The Shared Access Signature is required and should not be empty or blank and must be supplied as a string consisting of two parts in the format 'SharedAccessSignature sr=<resource_uri>&sig=<signature>&se=<expiry>' with an optional skn=<keyname>"
            )

        if len(parts) != 2:
            raise ValueError(
                "The Shared Access Signature must be of the format 'SharedAccessSignature sr=<resource_uri>&sig=<signature>&se=<expiry>' or/and it can additionally contain an optional skn=<keyname> name=value pair."
            )

        if len(sas_args) != len(d):
            raise ValueError("Invalid Shared Access Signature - Unable to parse")
        if not all(key in _valid_keys for key in d.keys()):
            raise ValueError(
                "Invalid keys in Shared Access Signature. The valid keys are sr, sig, se and an optional skn."
            )

        _validate_required_keys(d)

        try:
            unquoted_resource_uri = urllib.parse.unquote_plus(d.get(RESOURCE_URI))
            url_segments = unquoted_resource_uri.split(URI_SEPARATOR)

            module_id = None
            hostname = url_segments[0]
            device_id = url_segments[2]

            if len(url_segments) > 4:
                module_id = url_segments[4]

            return SharedAccessSignatureAuthenticationProvider(
                hostname, device_id, module_id, sas_token_str
            )
        except IndexError:
            raise ValueError(
                "One of the name value pair of the Shared Access Signature string should be a proper resource uri"
            )


def _validate_required_keys(d):
    """
    Validates that required keys are present.
    Raise ValueError if incorrect combination of keys
    """
    resource_uri = d.get(RESOURCE_URI)
    signature = d.get(SIGNATURE)
    expiry = d.get(EXPIRY)

    if resource_uri and signature and expiry:
        pass
    else:
        raise ValueError(
            "Invalid Shared Access Signature. It must be of the format 'SharedAccessSignature sr=<resource_uri>&sig=<signature>&se=<expiry>' or/and it can additionally contain an optional skn=<keyname> name=value pair."
        )



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

import os
import base64
import json
import six.moves.urllib as urllib
import requests
import requests_unixsocket

requests_unixsocket.monkeypatch()


class IotEdgeHsm(object):
    """
    Constructor for instantiating a iot hsm object.  This is an object that
    communicates with the Azure IoT Edge HSM in order to get connection credentials
    for an Azure IoT Edge module.  The credentials that this object return come in
    two forms:

    1. The trust bundle, which is a certificate that can be used as a trusted cert
       to authenticate the SSL connection between the IoE Edge module and IoT Edge
    2. A signing function, which can be used to create the sig field for a
       SharedAccessSignature string which can be used to authenticate with Iot Edge

    Instantiating this object does not require any parameters.  All necessary parameters
    come from environment variables that are set inside the IoT Edge module container
    by the edgeAgent that creates the module.
    """

    @staticmethod
    def _fix_socket_uri(old_uri):
        """
        This function takes a socket URI in one form and converts it into another form.

        The source form is based on what we receive inside the IOTEDGE_WORKLOADURI
        environment variable, and it looks like this:
        "unix:///var/run/iotedge/workload.sock"

        The destination form is based on what the requests_unixsocket library expects
        and it looks like this:
        "http+unix://%2Fvar%2Frun%2Fiotedge%2Fworkload.sock/"

        The function changes the prefix, uri-encodes the path, and adds a slash
        at the end.

        If the socket URI does not start with unix:// this function only adds
        a slash at the end.

        :param old_uri: The URI in IOTEDGE_WORKLOADURI form

        :return: The URI in requests_unixsocket form
        """
        old_prefix = "unix://"
        new_prefix = "http+unix://"

        if old_uri.startswith(old_prefix):
            stripped_uri = old_uri[len(old_prefix) :]
            if stripped_uri.endswith("/"):
                stripped_uri = stripped_uri[:-1]
            new_uri = new_prefix + urllib.parse.quote(stripped_uri, safe="")
        else:
            new_uri = old_uri

        if not new_uri.endswith("/"):
            new_uri += "/"

        return new_uri

    def __init__(self):
        """
        Constructor for instantiating a Azure IoT Edge HSM object
        """
        # All of these environment variables are required.  If any are missing,
        # we want this to fail.
        self.module_id = os.environ["IOTEDGE_MODULEID"]
        self.api_version = os.environ["IOTEDGE_APIVERSION"]
        self.module_generation_id = os.environ["IOTEDGE_MODULEGENERATIONID"]
        self.workload_uri = IotEdgeHsm._fix_socket_uri(os.environ["IOTEDGE_WORKLOADURI"])

    def get_trust_bundle(self):
        """
        Return the trust bundle that can be used to validate the server-side SSL
        TLS connection that we use to talk to edgeHub.

        :return: The CA certificate to use for connections to the Azure IoT Edge
        instance, as a PEM certificate in string form.
        """
        r = requests.get(
            self.workload_uri + "trust-bundle", params={"api-version": self.api_version}
        )
        r.raise_for_status()
        return r.json()["certificate"]

    def sign(self, data):
        """
        Use the IoTEdge HSM to sign a piece of data.  The caller should then insert the
        returned value (the signature) into the 'sig' field of a SharedAccessSignature string.

        :param data: The string to sign

        :return: The signature, as a URI-encoded and base64-encoded value that is ready to
        directly insert into the SharedAccessSignature string.
        """
        path = (
            self.workload_uri
            + "modules/"
            + urllib.parse.quote(self.module_id)
            + "/genid/"
            + self.module_generation_id
            + "/sign"
        )
        sign_request = {
            "keyId": "primary",
            "algo": "HMACSHA256",
            "data": base64.b64encode(data.encode("utf-8")).decode(),
        }

        r = requests.post(
            path, params={"api-version": self.api_version}, data=json.dumps(sign_request)
        )
        r.raise_for_status()
        return urllib.parse.quote(r.json()["digest"])



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""This module contains a class representing messages that are sent or received.
"""

import logging

logger = logging.getLogger(__name__)


class Message(object):
    """Represents a message to or from IoTHub

    :ivar data: The data that constitutes the payload
    :ivar custom_properties: Dictionary of custom message properties
    :ivar lock_token: Used by receiver to abandon, reject or complete the message
    :ivar message id: A user-settlable identifier for the message used for request-reply patterns. Format: A case-sensitive string (up to 128 characters long) of ASCII 7-bit alphanumeric characters + {'-', ':', '.', '+', '%', '_', '#', '*', '?', '!', '(', ')', ',', '=', '@', ';', '$', '''}
    :ivar sequence_number: A number (unique per device-queue) assigned by IoT Hub to each message
    :ivar to: A destination specified for Cloud-to-Device (C2D) messages
    :ivar expiry_time_utc: Date and time of message expiration in UTC format
    :ivar enqueued_time: Date and time a C2D message was received by IoT Hub
    :ivar correlation_id: A property in a response message that typically contains the message_id of the request, in request-reply patterns
    :ivar user_id: An ID to specify the origin of messages
    :ivar ack: A feedback message generator. This property is used in C2D messages to request IoT Hub to generate feedback messages as a result of the consumption of the message by the device
    :ivar content_encoding: Content encoding of the message data. Can be 'utf-8', 'utf-16' or 'utf-32'
    :ivar content_type: Content type property used to route messages with the message-body. Can be 'application/json'
    """

    def __init__(self, data, message_id=None, content_encoding=None, content_type=None):
        """
        Initializer for Message

        :param data: The  data that constitutes the payload
        :param message_id: A user-settable identifier for the message used for request-reply patterns. Format: A case-sensitive string (up to 128 characters long) of ASCII 7-bit alphanumeric characters + {'-', ':', '.', '+', '%', '_', '#', '*', '?', '!', '(', ')', ',', '=', '@', ';', '$', '''}
        :param content_encoding: Content encoding of the message data. Can be 'utf-8', 'utf-16' or 'utf-32'
        :param content_type: Content type property used to routes with the message body. Can be 'application/json'
        """
        self.data = data
        self.custom_properties = {}
        self.lock_token = None
        self.message_id = message_id
        self.sequence_number = None
        self.to = None
        self.expiry_time_utc = None
        self.enqueued_time = None
        self.correlation_id = None
        self.user_id = None
        self.ack = None
        self.content_encoding = content_encoding
        self.content_type = content_type
        self.output_name = None



"""Azure IoT Hub Device SDK Authentication

This package provides authentication-related functionality for use with the
Azure IoT Hub Device SDK.
"""

from .authentication_provider_factory import (
    from_connection_string,
    from_shared_access_signature,
    from_environment,
)

__all__ = ["from_connection_string", "from_shared_access_signature", "from_environment"]



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

import abc
import six
from . import constant


@six.add_metaclass(abc.ABCMeta)
class AbstractTransport:
    """
    All specific transport will follow implementations of this abstract class.
    """

    def __init__(self, auth_provider):
        self._auth_provider = auth_provider
        self.feature_enabled = {constant.C2D_MSG: False, constant.INPUT_MSG: False}

        # Event Handlers - Will be set by Client after instantiation of Transport
        self.on_transport_connected = None
        self.on_transport_disconnected = None
        self.on_transport_c2d_message_received = None
        self.on_transport_input_message_received = None
        self.on_transport_method_request_received = None

    @abc.abstractmethod
    def connect(self, callback):
        """
        Connect to the specific messaging system used by the specific transport protocol
        """
        pass

    @abc.abstractmethod
    def disconnect(self, callback):
        """
        Disconnect from the specific messaging system used by the specific transport protocol
        """
        pass

    @abc.abstractmethod
    def enable_feature(self, feature_name, callback=None):
        """
        Enable a specific feature (c2d, input, etc.)
        """
        pass

    @abc.abstractmethod
    def disable_feature(self, feature_name, callback=None):
        """
        Disable a specific feature (c2d, input, etc.)
        """
        pass

    @abc.abstractmethod
    def send_event(self, event, callback):
        """
        Send some telemetry, event or message.
        """
        pass

    @abc.abstractmethod
    def send_output_event(self, event, callback):
        """
        Send some event or message to a specific output
        """
        pass

    # TODO: consider changing this signature (should the response already be packaged?)
    @abc.abstractmethod
    def send_method_response(self, method, payload, status, callback=None):
        """
        Send a method response.
        """
        pass



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""This module contains user-facing synchronous clients for the
Azure IoTHub Device SDK for Python.
"""

import logging
from threading import Event
from .abstract_clients import (
    AbstractIoTHubClient,
    AbstractIoTHubDeviceClient,
    AbstractIoTHubModuleClient,
)
from .transport import constant
from .models import Message
from .inbox_manager import InboxManager
from .sync_inbox import SyncClientInbox

logger = logging.getLogger(__name__)


class GenericIoTHubClient(AbstractIoTHubClient):
    """A superclass representing a generic synchronous client.
    This class needs to be extended for specific clients.
    """

    def __init__(self, transport):
        """Initializer for a generic synchronous client.

        This initializer should not be called directly.
        Instead, the class method `from_authentication_provider` should be used to create a client object.

        :param transport: The transport that the client will use.
        """
        super(GenericIoTHubClient, self).__init__(transport)
        self._inbox_manager = InboxManager(inbox_type=SyncClientInbox)
        self._transport.on_transport_connected = self._on_state_change
        self._transport.on_transport_disconnected = self._on_state_change
        self._transport.on_transport_method_request_received = (
            self._inbox_manager.route_method_request
        )

    def _on_state_change(self, new_state):
        """Handler to be called by the transport upon a connection state change."""
        logger.info("Connection State - {}".format(new_state))

        if new_state == "disconnected":
            self._on_disconnected()

    def _on_disconnected(self):
        """Helper handler that is called upon a a transport disconnect"""
        self._inbox_manager.clear_all_method_requests()
        logger.info("Cleared all pending method requests due to disconnect")

    def connect(self):
        """Connects the client to an Azure IoT Hub or Azure IoT Edge Hub instance.

        The destination is chosen based on the credentials passed via the auth_provider parameter
        that was provided when this object was initialized.

        This is a synchronous call, meaning that this function will not return until the connection
        to the service has been completely established.
        """
        logger.info("Connecting to Hub...")

        connect_complete = Event()

        def callback():
            connect_complete.set()
            logger.info("Successfully connected to Hub")

        self._transport.connect(callback=callback)
        connect_complete.wait()

    def disconnect(self):
        """Disconnect the client from the Azure IoT Hub or Azure IoT Edge Hub instance.

        This is a synchronous call, meaning that this function will not return until the connection
        to the service has been completely closed.
        """
        logger.info("Disconnecting from Hub...")

        disconnect_complete = Event()

        def callback():
            disconnect_complete.set()
            logger.info("Successfully disconnected from Hub")

        self._transport.disconnect(callback=callback)
        disconnect_complete.wait()

    def send_event(self, message):
        """Sends a message to the default events endpoint on the Azure IoT Hub or Azure IoT Edge Hub instance.

        This is a synchronous event, meaning that this function will not return until the event
        has been sent to the service and the service has acknowledged receipt of the event.

        If the connection to the service has not previously been opened by a call to connect, this
        function will open the connection before sending the event.

        :param message: The actual message to send. Anything passed that is not an instance of the
        Message class will be converted to Message object.
        """
        if not isinstance(message, Message):
            message = Message(message)

        logger.info("Sending message to Hub...")
        send_complete = Event()

        def callback():
            send_complete.set()
            logger.info("Successfully sent message to Hub")

        self._transport.send_event(message, callback=callback)
        send_complete.wait()

    def receive_method_request(self, method_name=None, block=True, timeout=None):
        """Receive a method request via the Azure IoT Hub or Azure IoT Edge Hub.

        :param str method_name: Optionally provide the name of the method to receive requests for.
        If this parameter is not given, all methods not already being specifically targeted by
        a different request to receive_method will be received.
        :param bool block: Indicates if the operation should block until a request is received.
        Default True.
        :param int timeout: Optionally provide a number of seconds until blocking times out.

        :raises: InboxEmpty if timeout occurs on a blocking operation.
        :raises: InboxEmpty if no request is available on a non-blocking operation.

        :returns: MethodRequest object representing the received method request.
        """
        if not self._transport.feature_enabled[constant.METHODS]:
            self._enable_feature(constant.METHODS)

        method_inbox = self._inbox_manager.get_method_request_inbox(method_name)

        logger.info("Waiting for method request...")
        method_call = method_inbox.get(block=block, timeout=timeout)
        logger.info("Received method request")
        return method_call

    def send_method_response(self, method_request, payload, status):
        """Send a response to a method request via the Azure IoT Hub or Azure IoT Edge Hub.

        :param method_request: MethodRequest object representing the method request being
        responded to.
        :param payload: The desired payload for the method response.
        :param int status: The desired return status code for the method response.
        """
        logger.info("Sending method response to Hub...")
        send_complete = Event()

        def callback():
            send_complete.set()
            logger.info("Successfully sent method response to Hub")

        # TODO: maybe consolidate method_request, result and status into a new object
        self._transport.send_method_response(method_request, payload, status, callback=callback)
        send_complete.wait()

    def _enable_feature(self, feature_name):
        """Enable an Azure IoT Hub feature in the transport.

        This is a synchronous call, meaning that this function will not return until the feature
        has been enabled.

        :param feature_name: The name of the feature to enable.
        See azure.iot.device.iothub.transport.constant for possible values
        """
        logger.info("Enabling feature:" + feature_name + "...")
        enable_complete = Event()

        def callback():
            enable_complete.set()
            logger.info("Successfully enabled feature:" + feature_name)

        self._transport.enable_feature(feature_name, callback=callback)
        enable_complete.wait()


class IoTHubDeviceClient(GenericIoTHubClient, AbstractIoTHubDeviceClient):
    """A synchronous device client that connects to an Azure IoT Hub instance.

    Intended for usage with Python 2.7 or compatibility scenarios for Python 3.5.3+.
    """

    def __init__(self, transport):
        """Initializer for a IoTHubDeviceClient.

        This initializer should not be called directly.
        Instead, the class method `from_authentication_provider` should be used to create a client object.

        :param transport: The transport that the client will use.
        """
        super(IoTHubDeviceClient, self).__init__(transport)
        self._transport.on_transport_c2d_message_received = self._inbox_manager.route_c2d_message

    def receive_c2d_message(self, block=True, timeout=None):
        """Receive a C2D message that has been sent from the Azure IoT Hub.

        :param bool block: Indicates if the operation should block until a message is received.
        Default True.
        :param int timeout: Optionally provide a number of seconds until blocking times out.

        :raises: InboxEmpty if timeout occurs on a blocking operation.
        :raises: InboxEmpty if no message is available on a non-blocking operation.

        :returns: Message that was sent from the Azure IoT Hub.
        """
        if not self._transport.feature_enabled[constant.C2D_MSG]:
            self._enable_feature(constant.C2D_MSG)
        c2d_inbox = self._inbox_manager.get_c2d_message_inbox()

        logger.info("Waiting for C2D message...")
        message = c2d_inbox.get(block=block, timeout=timeout)
        logger.info("C2D message received")
        return message


class IoTHubModuleClient(GenericIoTHubClient, AbstractIoTHubModuleClient):
    """A synchronous module client that connects to an Azure IoT Hub or Azure IoT Edge instance.

    Intended for usage with Python 2.7 or compatibility scenarios for Python 3.5.3+.
    """

    def __init__(self, transport):
        """Intializer for a IoTHubModuleClient.

        This initializer should not be called directly.
        Instead, the class method `from_authentication_provider` should be used to create a client object.

        :param transport: The transport that the client will use.
        """
        super(IoTHubModuleClient, self).__init__(transport)
        self._transport.on_transport_input_message_received = (
            self._inbox_manager.route_input_message
        )

    def send_to_output(self, message, output_name):
        """Sends an event/message to the given module output.

        These are outgoing events and are meant to be "output events".

        This is a synchronous event, meaning that this function will not return until the event
        has been sent to the service and the service has acknowledged receipt of the event.

        If the connection to the service has not previously been opened by a call to connect, this
        function will open the connection before sending the event.

        :param message: message to send to the given output. Anything passed that is not an instance of the
        Message class will be converted to Message object.
        :param output_name: Name of the output to send the event to.
        """
        if not isinstance(message, Message):
            message = Message(message)
        message.output_name = output_name

        logger.info("Sending message to output:" + output_name + "...")
        send_complete = Event()

        def callback():
            logger.info("Successfully sent message to output: " + output_name)
            send_complete.set()

        self._transport.send_output_event(message, callback)
        send_complete.wait()

    def receive_input_message(self, input_name, block=True, timeout=None):
        """Receive an input message that has been sent from another Module to a specific input.

        :param str input_name: The input name to receive a message on.
        :param bool block: Indicates if the operation should block until a message is received.
        Default True.
        :param int timeout: Optionally provide a number of seconds until blocking times out.

        :raises: InboxEmpty if timeout occurs on a blocking operation.
        :raises: InboxEmpty if no message is available on a non-blocking operation.

        :returns: Message that was sent to the specified input.
        """
        if not self._transport.feature_enabled[constant.INPUT_MSG]:
            self._enable_feature(constant.INPUT_MSG)
        input_inbox = self._inbox_manager.get_input_message_inbox(input_name)

        logger.info("Waiting for input message on: " + input_name + "...")
        message = input_inbox.get(block=block, timeout=timeout)
        logger.info("Input message received on: " + input_name)
        return message



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""This module provides factory functions for creating authentication providers"""

from .sk_authentication_provider import SymmetricKeyAuthenticationProvider
from .sas_authentication_provider import SharedAccessSignatureAuthenticationProvider
from .iotedge_authentication_provider import IotEdgeAuthenticationProvider


def from_connection_string(connection_string):
    """Provides an AuthenticationProvider object that can be created simply with a connection string.
    :param connection_string: The connecting string.
    :return: a Symmetric Key AuthenticationProvider.
    """
    return SymmetricKeyAuthenticationProvider.parse(connection_string)


def from_shared_access_signature(sas_token_str):
    """Provides an `AuthenticationProvider` object that can be created simply with a shared access signature.
    :param sas_token_str: The shared access signature.
    :return: Shared Access Signature AuthenticationProvider.
    """
    return SharedAccessSignatureAuthenticationProvider.parse(sas_token_str)


def from_environment():
    """Provides an `AuthenticationProvider` object that can be used inside of an Azure IoT Edge module.

    This method does not need any parameters because all of the information necessary to connect
    to Azure IoT Edge comes from the operating system of the module container and also from the
    IoTEdge service.

    :return: iotedge AuthenticationProvider.
    """
    return IotEdgeAuthenticationProvider()



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

import os
import logging
from .base_renewable_token_authentication_provider import BaseRenewableTokenAuthenticationProvider
from .iotedge_hsm import IotEdgeHsm

logger = logging.getLogger(__name__)


class IotEdgeAuthenticationProvider(BaseRenewableTokenAuthenticationProvider):
    """An Azure IoT Edge Authentication Provider.

    This provider creates the Shared Access Signature that would be needed to connenct to the IoT Edge runtime
    """

    def __init__(self):
        """
        Constructor for IoT Edge Authentication Provider
        """
        hostname = os.environ["IOTEDGE_IOTHUBHOSTNAME"]
        device_id = os.environ["IOTEDGE_DEVICEID"]
        module_id = os.environ["IOTEDGE_MODULEID"]

        logger.info("Using IoTEdge authentication for {%s, %s, %s}", hostname, device_id, module_id)

        BaseRenewableTokenAuthenticationProvider.__init__(self, hostname, device_id, module_id)

        self.hsm = IotEdgeHsm()
        self.gateway_hostname = os.environ["IOTEDGE_GATEWAYHOSTNAME"]
        self.ca_cert = self.hsm.get_trust_bundle()

    @staticmethod
    def parse(connection_string):
        pass

    def _sign(self, quoted_resource_uri, expiry):
        """
        Creates the signature to be inserted in the SAS token
        :param resource_uri: the resource URI to encode into the token
        :param expiry: an integer value representing the number of seconds since the epoch 00:00:00 UTC on 1 January 1970 at which the token will expire.
        :return: The signature portion of the Sas Token.
        """
        string_to_sign = quoted_resource_uri + "\n" + str(expiry)
        return self.hsm.sign(string_to_sign)



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""This module contains tools for working with Shared Access Signature (SAS) Tokens"""

import base64
import hmac
import hashlib
import time
import six.moves.urllib as urllib

__all__ = ["SasToken", "SasTokenError"]


class SasTokenError(Exception):
    """Error in SasToken"""

    def __init__(self, message, cause=None):
        """Initializer for SasTokenError

        :param str message: Error message
        :param cause: Exception that caused this error (optional)
        """
        super(SasTokenError, self).__init__(message)
        self.cause = cause


class SasToken(object):
    """Shared Access Signature Token used to authenticate a request

    Parameters:
    uri (str): URI of the resouce to be accessed
    key_name (str): Shared Access Key Name
    key (str): Shared Access Key (base64 encoded)
    ttl (int)[default 3600]: Time to live for the token, in seconds

    Data Attributes:
    expiry_time (int): Time that token will expire (in UTC, since epoch)
    ttl (int): Time to live for the token, in seconds

    Raises:
    SasTokenError if trying to build a SasToken from invalid values
    """

    _encoding_type = "utf-8"
    _service_token_format = "SharedAccessSignature sr={}&sig={}&se={}&skn={}"
    _device_token_format = "SharedAccessSignature sr={}&sig={}&se={}"

    def __init__(self, uri, key, key_name=None, ttl=3600):
        self._uri = urllib.parse.quote_plus(uri)
        self._key = key
        self._key_name = key_name
        self.ttl = ttl
        self.refresh()

    def __repr__(self):
        return self._token

    def refresh(self):
        """
        Refresh the SasToken lifespan, giving it a new expiry time
        """
        self.expiry_time = int(time.time() + self.ttl)
        self._token = self._build_token()

    def _build_token(self):
        """Buid SasToken representation

        Returns:
        String representation of the token
        """
        try:
            message = (self._uri + "\n" + str(self.expiry_time)).encode(self._encoding_type)
            signing_key = base64.b64decode(self._key.encode(self._encoding_type))
            signed_hmac = hmac.HMAC(signing_key, message, hashlib.sha256)
            signature = urllib.parse.quote(base64.b64encode(signed_hmac.digest()))
        except (TypeError, base64.binascii.Error) as e:
            raise SasTokenError("Unable to build SasToken from given values", e)
        if self._key_name:
            token = self._service_token_format.format(
                self._uri, signature, str(self.expiry_time), self._key_name
            )
        else:
            token = self._device_token_format.format(self._uri, signature, str(self.expiry_time))
        return token



"""Azure IoT Hub Device Library

This library provides functionality for communicating with the Azure IoT Hub
as a Device or Module.
"""

from .sync_clients import IoTHubDeviceClient, IoTHubModuleClient
from .sync_inbox import InboxEmpty
from .models import Message

__all__ = ["IoTHubDeviceClient", "IoTHubModuleClient", "Message", "InboxEmpty"]



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

import paho.mqtt.client as mqtt
import logging
import ssl
import traceback

logger = logging.getLogger(__name__)


class MQTTProvider(object):
    """
    A wrapper over the actual implementation of mqtt message broker which will eventually connect to an mqtt broker
    to publish/subscribe messages.
    """

    def __init__(self, client_id, hostname, username, ca_cert=None):
        """
        Constructor to instantiate a mqtt provider.
        :param client_id: The id of the client connecting to the broker.
        :param hostname: hostname or IP address of the remote broker.
        :param ca_cert: Certificate which can be used to validate a server-side TLS connection.
        """
        self._client_id = client_id
        self._hostname = hostname
        self._username = username
        self._mqtt_client = None
        self._ca_cert = ca_cert

        self.on_mqtt_connected = None
        self.on_mqtt_disconnected = None
        self.on_mqtt_published = None
        self.on_mqtt_subscribed = None
        self.on_mqtt_unsubscribed = None
        self.on_mqtt_message_received = None

        self._create_mqtt_client()

    def _create_mqtt_client(self):
        """
        Create the MQTT client object and assign all necessary callbacks.
        """
        logger.info("creating mqtt client")

        self._mqtt_client = mqtt.Client(self._client_id, False, protocol=mqtt.MQTTv311)

        def on_connect_callback(client, userdata, flags, result_code):
            logger.info("connected with result code: %s", str(result_code))
            # TODO: how to do failed connection?
            try:
                self.on_mqtt_connected()
            except:  # noqa: E722 do not use bare 'except'
                logger.error("Unexpected error calling on_mqtt_connected")
                logger.error(traceback.format_exc())

        def on_disconnect_callback(client, userdata, result_code):
            logger.info("disconnected with result code: %s", str(result_code))
            try:
                self.on_mqtt_disconnected()
            except:  # noqa: E722 do not use bare 'except'
                logger.error("Unexpected error calling on_mqtt_disconnected")
                logger.error(traceback.format_exc())

        def on_publish_callback(client, userdata, mid):
            logger.info("payload published for %s", str(mid))
            # TODO: how to do failed publish
            try:
                self.on_mqtt_published(mid)
            except:  # noqa: E722 do not use bare 'except'
                logger.error("Unexpected error calling on_mqtt_published")
                logger.error(traceback.format_exc())

        def on_subscribe_callback(client, userdata, mid, granted_qos):
            logger.info("suback received for %s", str(mid))
            # TODO: how to do failure?
            try:
                self.on_mqtt_subscribed(mid)
            except:  # noqa: E722 do not use bare 'except'
                logger.error("Unexpected error calling on_mqtt_subscribed")
                logger.error(traceback.format_exc())

        def on_message_callback(client, userdata, mqtt_message):
            logger.info("message received on %s", mqtt_message.topic)
            try:
                self.on_mqtt_message_received(mqtt_message._topic, mqtt_message.payload)
            except:  # noqa: E722 do not use bare 'except'
                logger.error("Unexpected error calling on_mqtt_message_received")
                logger.error(traceback.format_exc())

        def on_unsubscribe_callback(client, userdata, mid):
            logger.info("UNSUBACK received for %s", str(mid))
            # TODO: how to do failure?
            try:
                self.on_mqtt_unsubscribed(mid)
            except:  # noqa: E722 do not use bare 'except'
                logger.error("Unexpected error calling on_mqtt_unsubscribed")
                logger.error(traceback.format_exc())

        self._mqtt_client.on_connect = on_connect_callback
        self._mqtt_client.on_disconnect = on_disconnect_callback
        self._mqtt_client.on_publish = on_publish_callback
        self._mqtt_client.on_subscribe = on_subscribe_callback
        self._mqtt_client.on_message = on_message_callback
        self._mqtt_client.on_unsubscribe = on_unsubscribe_callback

        logger.info("Created MQTT provider, assigned callbacks")

    def connect(self, password):
        """
        This method connects the upper transport layer to the mqtt broker.
        This method should be called as an entry point before sending any telemetry.
        """
        logger.info("connecting to mqtt broker")

        ssl_context = ssl.SSLContext(ssl.PROTOCOL_TLSv1_2)
        if self._ca_cert:
            ssl_context.load_verify_locations(cadata=self._ca_cert)
        else:
            ssl_context.load_default_certs()
        ssl_context.verify_mode = ssl.CERT_REQUIRED
        ssl_context.check_hostname = True
        self._mqtt_client.tls_set_context(ssl_context)
        self._mqtt_client.tls_insecure_set(False)
        self._mqtt_client.username_pw_set(username=self._username, password=password)

        self._mqtt_client.connect(host=self._hostname, port=8883)
        self._mqtt_client.loop_start()

    def reconnect(self, password):
        """
        This method reconnects the mqtt broker, possibly because of a password (sas) change
        Connect should have previously been called.
        """
        logger.info("reconnecting transport")
        self._mqtt_client.username_pw_set(username=self._username, password=password)
        self._mqtt_client.reconnect()

    def disconnect(self):
        """
        This method disconnects the mqtt provider. This should be called from the upper transport
        when it wants to disconnect from the mqtt provider.
        """
        logger.info("disconnecting transport")
        self._mqtt_client.disconnect()

    def publish(self, topic, message_payload):
        """
        This method enables the transport to send a message to the message broker.
        By default the the quality of service level to use is set to 1.
        :param topic: topic: The topic that the message should be published on.
        :param message_payload: The actual message to send.
        :return message ID for the publish request.
        """
        logger.info("sending")
        message_info = self._mqtt_client.publish(topic=topic, payload=message_payload, qos=1)
        return message_info.mid

    def subscribe(self, topic, qos=0):
        """
        This method subscribes the client to one topic.
        :param topic: a single string specifying the subscription topic to subscribe to
        :param qos: the desired quality of service level for the subscription. Defaults to 0.
        :return: message ID for the subscribe request
        Raises a ValueError if qos is not 0, 1 or 2, or if topic is None or has zero string length,
        """
        logger.info("subscribing to %s with qos %s", topic, str(qos))
        (result, mid) = self._mqtt_client.subscribe(topic, qos)
        return mid

    def unsubscribe(self, topic):
        """
        Unsubscribe the client from one topic.
        :param topic: a single string which is the subscription topic to unsubscribe from.
        :return: mid the message ID for the unsubscribe request.
        Raises a ValueError if topic is None or has zero string length, or is not a string.
        """
        logger.info("unsubscribing from %s", topic)
        (result, mid) = self._mqtt_client.unsubscribe(topic)
        return mid



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""This module contains abstract classes for the various clients of the Azure IoT Hub Device SDK
"""

import six
import abc
import logging
from .transport import MQTTTransport

logger = logging.getLogger(__name__)


@six.add_metaclass(abc.ABCMeta)
class AbstractIoTHubClient(object):
    """A superclass representing a generic client. This class needs to be extended for specific clients."""

    def __init__(self, transport):
        """Initializer for a generic client.

        :param transport: The transport that the client will use.
        """
        self._transport = transport

    @classmethod
    def from_authentication_provider(cls, authentication_provider, transport_name):
        """Creates a client with the specified authentication provider and transport.

        When creating the client, you need to pass in an authorization provider and a transport_name.

        The authentication_provider parameter is an object created using the authentication_provider_factory
        module.  It knows where to connect (a network address), how to authenticate with the service
        (a set of credentials), and, if necessary, the protocol gateway to use when communicating
        with the service.

        The transport_name is a string which defines the name of the transport to use when connecting
        with the service or the protocol gateway.

        Currently "mqtt" is the only supported transport.

        :param authentication_provider: The authentication provider.
        :param transport_name: The name of the transport that the client will use.

        :returns: Instance of the client.

        :raises: ValueError if given an invalid transport_name.
        :raises: NotImplementedError if transport_name is "amqp" or "http".
        """
        transport_name = transport_name.lower()
        if transport_name == "mqtt":
            transport = MQTTTransport(authentication_provider)
        elif transport_name == "amqp" or transport_name == "http":
            raise NotImplementedError("This transport has not yet been implemented")
        else:
            raise ValueError("No specific transport can be instantiated based on the choice.")
        return cls(transport)

    @abc.abstractmethod
    def connect(self):
        pass

    @abc.abstractmethod
    def disconnect(self):
        pass

    @abc.abstractmethod
    def send_event(self, message):
        pass

    @abc.abstractmethod
    def receive_method_request(self, method_name=None):
        pass

    @abc.abstractmethod
    def send_method_response(self, method_request, payload, status):
        pass


@six.add_metaclass(abc.ABCMeta)
class AbstractIoTHubDeviceClient(AbstractIoTHubClient):
    @abc.abstractmethod
    def receive_c2d_message(self):
        pass


@six.add_metaclass(abc.ABCMeta)
class AbstractIoTHubModuleClient(AbstractIoTHubClient):
    @abc.abstractmethod
    def send_to_output(self, message, output_name):
        pass

    @abc.abstractmethod
    def receive_input_message(self, input_name):
        pass



""" Azure IoT Device Library

This library provides clients and associated models for communicating with Azure IoT services
from an IoT device.
"""

from .iothub import *
from .iothub import auth  # Consider moving this to common after DPS added
from . import iothub

# iothub and common subpackages are still showing up in intellisense


__all__ = iothub.__all__  # + provisioning.__all__ etc.



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""This module contains user-facing asynchronous clients for the
Azure IoTHub Device SDK for Python.
"""

import logging
from azure.iot.device.common import async_adapter
from azure.iot.device.iothub.abstract_clients import (
    AbstractIoTHubClient,
    AbstractIoTHubDeviceClient,
    AbstractIoTHubModuleClient,
)
from azure.iot.device.iothub.models import Message
from azure.iot.device.iothub.transport import constant
from azure.iot.device.iothub.inbox_manager import InboxManager
from .async_inbox import AsyncClientInbox

logger = logging.getLogger(__name__)


class GenericIoTHubClient(AbstractIoTHubClient):
    """A super class representing a generic asynchronous client.
    This class needs to be extended for specific clients.
    """

    def __init__(self, transport):
        """Initializer for a generic asynchronous client.

        This initializer should not be called directly.
        Instead, the class method `from_authentication_provider` should be used to create a client object.

        :param transport: The transport that the client will use.
        """
        super().__init__(transport)
        self._inbox_manager = InboxManager(inbox_type=AsyncClientInbox)
        self._transport.on_transport_connected = self._on_state_change
        self._transport.on_transport_disconnected = self._on_state_change
        self._transport.on_transport_method_request_received = (
            self._inbox_manager.route_method_request
        )

    def _on_state_change(self, new_state):
        """Handler to be called by the transport upon a connection state change."""
        logger.info("Connection State - {}".format(new_state))

        if new_state == "disconnected":
            self._on_disconnected()

    def _on_disconnected(self):
        """Helper handler that is called upon a a transport disconnect"""
        self._inbox_manager.clear_all_method_requests()
        logger.info("Cleared all pending method requests due to disconnect")

    async def connect(self):
        """Connects the client to an Azure IoT Hub or Azure IoT Edge Hub instance.

        The destination is chosen based on the credentials passed via the auth_provider parameter
        that was provided when this object was initialized.
        """
        logger.info("Connecting to Hub...")
        connect_async = async_adapter.emulate_async(self._transport.connect)

        def sync_callback():
            logger.info("Successfully connected to Hub")

        callback = async_adapter.AwaitableCallback(sync_callback)

        await connect_async(callback=callback)
        await callback.completion()

    async def disconnect(self):
        """Disconnect the client from the Azure IoT Hub or Azure IoT Edge Hub instance.
        """
        logger.info("Disconnecting from Hub...")
        disconnect_async = async_adapter.emulate_async(self._transport.disconnect)

        def sync_callback():
            logger.info("Successfully disconnected from Hub")

        callback = async_adapter.AwaitableCallback(sync_callback)

        await disconnect_async(callback=callback)
        await callback.completion()

    async def send_event(self, message):
        """Sends a message to the default events endpoint on the Azure IoT Hub or Azure IoT Edge Hub instance.

        If the connection to the service has not previously been opened by a call to connect, this
        function will open the connection before sending the event.

        :param message: The actual message to send. Anything passed that is not an instance of the
        Message class will be converted to Message object.
        """
        if not isinstance(message, Message):
            message = Message(message)

        logger.info("Sending message to Hub...")
        send_event_async = async_adapter.emulate_async(self._transport.send_event)

        def sync_callback():
            logger.info("Successfully sent message to Hub")

        callback = async_adapter.AwaitableCallback(sync_callback)

        await send_event_async(message, callback=callback)
        await callback.completion()

    async def receive_method_request(self, method_name=None):
        """Receive a method request via the Azure IoT Hub or Azure IoT Edge Hub.

        If no method request is yet available, will wait until it is available.

        :param str method_name: Optionally provide the name of the method to receive requests for.
        If this parameter is not given, all methods not already being specifically targeted by
        a different call to receive_method will be received.

        :returns: MethodRequest object representing the received method request.
        """
        if not self._transport.feature_enabled[constant.METHODS]:
            await self._enable_feature(constant.METHODS)

        method_inbox = self._inbox_manager.get_method_request_inbox(method_name)

        logger.info("Waiting for method request...")
        method_request = await method_inbox.get()
        logger.info("Received method request")
        return method_request

    async def send_method_response(self, method_request, payload, status):
        """Send a response to a method request via the Azure IoT Hub or Azure IoT Edge Hub.

        :param method_request: MethodRequest object representing the method request being
        responded to.
        :param payload: The desired payload for the method response.
        :param int status: The desired return status code for the method response.
        """
        logger.info("Sending method response to Hub...")
        send_method_response_async = async_adapter.emulate_async(
            self._transport.send_method_response
        )

        def sync_callback():
            logger.info("Successfully sent method response to Hub")

        callback = async_adapter.AwaitableCallback(sync_callback)

        # TODO: maybe consolidate method_request, result and status into a new object
        await send_method_response_async(method_request, payload, status, callback=callback)
        await callback.completion()

    async def _enable_feature(self, feature_name):
        """Enable an Azure IoT Hub feature in the transport

        :param feature_name: The name of the feature to enable.
        See azure.iot.device.iothub.transport.constant for possible values.
        """
        logger.info("Enabling feature:" + feature_name + "...")
        enable_feature_async = async_adapter.emulate_async(self._transport.enable_feature)

        def sync_callback():
            logger.info("Successfully enabled feature:" + feature_name)

        callback = async_adapter.AwaitableCallback(sync_callback)

        await enable_feature_async(feature_name, callback=callback)


class IoTHubDeviceClient(GenericIoTHubClient, AbstractIoTHubDeviceClient):
    """An asynchronous device client that connects to an Azure IoT Hub instance.

    Intended for usage with Python 3.5.3+
    """

    def __init__(self, transport):
        super().__init__(transport)
        self._transport.on_transport_c2d_message_received = self._inbox_manager.route_c2d_message

    async def receive_c2d_message(self):
        """Receive a C2D message that has been sent from the Azure IoT Hub.

        If no message is yet available, will wait until an item is available.

        :returns: Message that was sent from the Azure IoT Hub.
        """
        if not self._transport.feature_enabled[constant.C2D_MSG]:
            await self._enable_feature(constant.C2D_MSG)
        c2d_inbox = self._inbox_manager.get_c2d_message_inbox()

        logger.info("Waiting for C2D message...")
        message = await c2d_inbox.get()
        logger.info("C2D message received")
        return message


class IoTHubModuleClient(GenericIoTHubClient, AbstractIoTHubModuleClient):
    """An asynchronous module client that connects to an Azure IoT Hub or Azure IoT Edge instance.

    Intended for usage with Python 3.5.3+
    """

    def __init__(self, transport):
        super().__init__(transport)
        self._transport.on_transport_input_message_received = (
            self._inbox_manager.route_input_message
        )

    async def send_to_output(self, message, output_name):
        """Sends an event/message to the given module output.

        These are outgoing events and are meant to be "output events"

        If the connection to the service has not previously been opened by a call to connect, this
        function will open the connection before sending the event.

        :param message: message to send to the given output. Anything passed that is not an instance of the
        Message class will be converted to Message object.
        :param output_name: Name of the output to send the event to.
        """
        if not isinstance(message, Message):
            message = Message(message)

        message.output_name = output_name

        logger.info("Sending message to output:" + output_name + "...")
        send_output_event_async = async_adapter.emulate_async(self._transport.send_output_event)

        def sync_callback():
            logger.info("Successfully sent message to output: " + output_name)

        callback = async_adapter.AwaitableCallback(sync_callback)

        await send_output_event_async(message, callback)
        await callback.completion()

    async def receive_input_message(self, input_name):
        """Receive an input message that has been sent from another Module to a specific input.

        If no message is yet available, will wait until an item is available.

        :param str input_name: The input name to receive a message on.
        :returns: Message that was sent to the specified input.
        """
        if not self._transport.feature_enabled[constant.INPUT_MSG]:
            await self._enable_feature(constant.INPUT_MSG)
        inbox = self._inbox_manager.get_input_message_inbox(input_name)

        logger.info("Waiting for input message on: " + input_name + "...")
        message = await inbox.get()
        logger.info("Input message received on: " + input_name)
        return message



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""This module contains a class representing a request to invoke a direct method.
"""


class MethodRequest(object):
    """Represents a request to invoke a direct method.

    :ivar str name: The name of the method to be invoked
    :ivar payload: The payload being sent with the request.
    """

    def __init__(self, request_id, name, payload):
        """Initializer for a MethodRequest.

        :param str request_id: The request id of the request.
        :param str name: The name of the method to be invoked
        :param payload: The payload being sent with the request.
        """
        self._request_id = request_id
        self._name = name
        self._payload = payload

    @property
    def name(self):
        return self._name

    @property
    def payload(self):
        return self._payload



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""This module contains an Inbox class for use with an asynchronous client"""

import janus
from azure.iot.device.iothub.sync_inbox import AbstractInbox


class AsyncClientInbox(AbstractInbox):
    """Holds generic incoming data for an asynchronous client.

    All methods implemented in this class are threadsafe.
    """

    def __init__(self):
        """Initializer for AsyncClientInbox."""
        self._queue = janus.Queue()

    def _put(self, item):
        """Put an item into the Inbox.

        Block if necessary until a free slot is available.
        Only to be used by the InboxManager.

        :param item: The item to be put in the Inbox.
        """
        self._queue.sync_q.put(item)

    async def get(self):
        """Remove and return an item from the Inbox.

        If Inbox is empty, wait until an item is available.

        :returns: An item from the Inbox.
        """
        return await self._queue.async_q.get()

    def empty(self):
        """Returns True if the inbox is empty, False otherwise

        :returns: Boolean indicating if the inbox is empty
        """
        return self._queue.async_q.empty()

    def clear(self):
        """Remove all items from the inbox.
        """
        while True:
            try:
                self._queue.sync_q.get_nowait()
            except janus.SyncQueueEmpty:
                break



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

from setuptools import setup, find_packages

# azure v0.x is not compatible with this package
# azure v0.x used to have a __version__ attribute (newer versions don't)
try:
    import azure

    try:
        ver = azure.__version__
        raise Exception(
            "This package is incompatible with azure=={}. ".format(ver)
            + 'Uninstall it with "pip uninstall azure".'
        )
    except AttributeError:
        pass
except ImportError:
    pass


with open("README.md", "r") as fh:
    _long_description = fh.read()


setup(
    name="azure-iot-device",
    version="2.0.0-preview.2",
    description="Microsoft Azure IoT Device Library",
    license="MIT License",
    url="https://github.com/Azure/azure-iot-sdk-python-preview",
    author="Microsoft Corporation",
    author_email="opensource@microsoft.com",
    long_description=_long_description,
    long_description_content_type="text/markdown",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
    ],
    install_requires=[
        "six>=1.12.0,<2.0.0",
        "paho-mqtt>=1.4.0,<2.0.0",
        "transitions>=0.6.8,<1.0.0",
        "requests>=2.20.0,<3.0.0",
        "requests-unixsocket>=0.1.5,<1.0.0",
        "janus>=0.4.0,<1.0.0;python_version>='3.5'",
    ],
    extras_require={":python_version<'3.0'": ["azure-iot-nspkg"]},
    python_requires=">=2.7, !=3.0.*, !=3.1.*, !=3.2.*, !=3.3*, <4",
    packages=find_packages(
        exclude=[
            "tests",
            "samples",
            # Exclude packages that will be covered by PEP420 or nspkg
            "azure",
            "azure.iot",
        ]
    ),
    zip_safe=False,
)
